local WMS = WMS_016
local REDIR_TIMEOUT_SEC = 3*60
-- Redirections are stored under a key: receiver..message.
-- Redirections selfestruct after a timeout.
-- They cumulate the senders of redirection notifications
function WMS:GetRedirections()
    if not self.redirections then
        self.redirections = {}
    end
    return self.redirections
end

function WMS:AddRedirection(from, message, to)
    local redirKey = self:CreateRedirectionKey(to, message)
    local redir = nil
    if (not self:GetRedirections()[redirKey]) then
        redir = {}
        redir.from = from
        redir.message = message
        redir.to = to
        redir.redirectors = {}
        
        redir.selfDestructTimer = self.CreateTimer(REDIR_TIMEOUT_SEC, 
            function ()
                self:GetRedirections()[redirKey] = nil
                self.redirectionCount = self.redirectionCount - 1
                if self.redirectionCount == 0 then
                    WMS:RedirectedDisplayRecord()
                    PlaySound("igMiniMapClose", "sfx")
                end
                WMS:PrintDebug ("expired: ",
                redir.from, " -> ", redir.to, ": ",
                redir.message)
            end
        )
        
        self:GetRedirections()[redirKey] = redir
        if not self.redirectionCount then 
            self.redirectionCount = 0
        end
        self.redirectionCount = self.redirectionCount + 1
        WMS:PrintDebug ("Redir added with key ", redirKey)
        redir.selfDestructTimer:Start();
        self:StartRedirectorTimeoutAnim(REDIR_TIMEOUT_SEC)
    end
    return redir
end

function WMS:CreateRedirectionKey(to, message)
    local redirKey = ""

--~     Sometimes the name is provided including the realm (Jack-Aegwyn) and 
--~     sometimes without (Jack). The plain name is sufficiant for use as key.    
    local endOfPrefix = string.find(to, "-")
    if endOfPrefix then 
        redirKey = string.sub(to, 1, endOfPrefix - 1)
    else
        redirKey = tostring(to)
    end
    
    redirKey = redirKey..tostring(message)
    return redirKey
end

function WMS:SendRedirectionNotification(redirection, redirector)
    if redirection and redirection.from then 
        local msg = redirector .. "\t".. redirection.message
        WMS:PrintDebug("Notification", UnitName("PLAYER"), "->", redirection.from, msg)
        self.COM:SendMsg(
            self.MSG_CODES.NOTIFY_REDIRECTED, 
            msg, 
            redirection.from
        )
    end
end

function WMS:RedirectedCheck(lastReceivedMessage, message)
    local from = nil
    if message and lastReceivedMessage and 
       (message == lastReceivedMessage.message) 
    then
        from = lastReceivedMessage.sender
    end

    local redirection = self:AddRedirection(
        from, 
        message, 
        self:GetReceiver()
    )

    if (from) then
        local redirector = UnitName("PLAYER")
        self:SendRedirectionNotification(redirection, redirector)
        self:SetLastReceivedMessage(nil, nil)
    end
end

function WMS:AddRedirector(redirection, redirector)
    for i, knownRedirector in ipairs(redirection.redirectors) do
        if redirector == knownRedirector then
            return false
        end
    end
    
    table.insert(redirection.redirectors, redirector)
    return true
end

function WMS:RedirectionNotified(rawMessage, sender)
--~     Extract the redirector from the raw message
    local endOfPrefix, startOfBody = string.find(rawMessage, "\t")
    local redirector = string.sub(rawMessage, 1, endOfPrefix - 1)
    local msgBody = string.sub(rawMessage, startOfBody + 1)
    
--~     Earlier, when we redirected the message, the key was created
--~     by the destination address and the body. Now the origin (sender)
--~     must be used to build the key
    local key = self:CreateRedirectionKey(sender, msgBody)    
    local redir = self:GetRedirections()[key];
    if (redir) then
        local redirectorIsNew = self:AddRedirector(redir, redirector)
        if (redirectorIsNew)then
            self:SendRedirectionNotification(redir, redirector)
            redir.selfDestructTimer:Start()
        else 
            WMS:PrintDebug ("Cycle detected, ", redirector, " already listed")
        end
        
        local printStr = ""
        for key, redir in ipairs(redir.redirectors) do
            if (printStr ~= "") then
                printStr = printStr .. ", "
            end
            printStr = printStr .. ", " .. tostring(redir)
        end
        WMS:PrintDebug ("known redirectors: ", printStr)

        local newHighscore = false
        while (self:GetConCount(WMS.MSG_CODES.NOTIFY_REDIRECTED) < #redir.redirectors) do
            self:IncConCount(WMS.MSG_CODES.NOTIFY_REDIRECTED)
            newHighscore = true;
        end

        self:GetRedirectorFontString():SetText(
            #redir.redirectors .." (" 
            .. TL("Highscore") .. ": " .. self:GetConCount(WMS.MSG_CODES.NOTIFY_REDIRECTED)
            .. ")\n" .. redirector
        )
        self:StartRedirectorFlashAnim()
        if newHighscore then
            PlaySoundFile("Sound\\INTERFACE\\AutoQuestComplete.ogg", "sfx")
        else
            PlaySoundFile("Sound\\INTERFACE\\Aggro_Pulled_Aggro.ogg", "sfx")
        end
    else        
--~         Should be unreachable, because a redir object is created for each new message
        WMS:PrintDebug ("No rediretion found for ", key)
    end
end

function WMS:RedirectedDisplayRecord()
    self:GetRedirectorFontString():SetText(
        TL("Highscore") .. ": " 
        .. WMS:GetConCount(WMS.MSG_CODES.NOTIFY_REDIRECTED)
    )
end

function WMS.RedirectedOnShow()
    WMS:SetRedirectionTextTilt()
    WMS:RedirectedDisplayRecord()
end